{% extends "core.html" %}

{% block main_pane %}
<div id="main-pane" class="col-xs-12" style="height: {{frame_height}}px; background-color: #dff0d8; padding: 30px; overflow:scroll;">
    <h1>Live Chat</h1>
    <hr style="border-top: 1px solid #555" />
    <span id="task-description" style="font-size: 16px">
    </span>
</div>
{% endblock %}